/******************************************************************************
 * @file     startup_SAMA5D2.c
 * @brief    CMSIS Startup File
 * @version  V1.00
 * @date     22 Feb 2017
 *
 * @note
 *
 ******************************************************************************/
/*
 * Copyright (c) 2009-2017 ARM Limited. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "RTE_Components.h"
#include CMSIS_device_header

/*----------------------------------------------------------------------------
  Definitions
 *----------------------------------------------------------------------------*/
#define USR_MODE 0x10            // User mode
#define FIQ_MODE 0x11            // Fast Interrupt Request mode
#define IRQ_MODE 0x12            // Interrupt Request mode
#define SVC_MODE 0x13            // Supervisor mode
#define ABT_MODE 0x17            // Abort mode
#define UND_MODE 0x1B            // Undefined Instruction mode
#define SYS_MODE 0x1F            // System mode

/*----------------------------------------------------------------------------
  Internal References
 *----------------------------------------------------------------------------*/
void Vectors       (void) __attribute__ ((section("RESET")));
void Reset_Handler(void);

/*----------------------------------------------------------------------------
  Exception / Interrupt Handler
 *----------------------------------------------------------------------------*/
void Undef_Handler (void) __attribute__ ((weak, alias("Default_Handler")));
void SVC_Handler   (void) __attribute__ ((weak, alias("Default_Handler")));
void PAbt_Handler  (void) __attribute__ ((weak, alias("Default_Handler")));
void DAbt_Handler  (void) __attribute__ ((weak, alias("Default_Handler")));

/*----------------------------------------------------------------------------
  Exception / Interrupt Vector Table
 *----------------------------------------------------------------------------*/
__ASM void Vectors(void) {
  IMPORT Undef_Handler
  IMPORT SVC_Handler
  IMPORT PAbt_Handler
  IMPORT DAbt_Handler
  LDR    PC, =Reset_Handler
  LDR    PC, =Undef_Handler
  LDR    PC, =SVC_Handler
  LDR    PC, =PAbt_Handler
  LDR    PC, =DAbt_Handler
  NOP
  LDR    PC, =IRQ_Handler
  LDR    PC, =FIQ_Handler
}

/*----------------------------------------------------------------------------
  Reset Handler called on controller reset
 *----------------------------------------------------------------------------*/
__ASM void Reset_Handler(void) {

  // Mask interrupts
  CPSID  if                           

  // Put any cores other than 0 to sleep
  MRC    p15, 0, R0, c0, c0, 5       // Read MPIDR
  ANDS   R0, R0, #3
goToSleep
  WFINE
  BNE    goToSleep

  // Reset SCTLR Settings
  MRC    p15, 0, R0, c1, c0, 0       // Read CP15 System Control register
  BIC    R0, R0, #(0x1 << 12)        // Clear I bit 12 to disable I Cache
  BIC    R0, R0, #(0x1 <<  2)        // Clear C bit  2 to disable D Cache
  BIC    R0, R0, #0x1                // Clear M bit  0 to disable MMU
  BIC    R0, R0, #(0x1 << 11)        // Clear Z bit 11 to disable branch prediction
  BIC    R0, R0, #(0x1 << 13)        // Clear V bit 13 to disable hivecs
  MCR    p15, 0, R0, c1, c0, 0       // Write value back to CP15 System Control register
  ISB

  // Configure ACTLR
  MRC    p15, 0, r0, c1, c0, 1       // Read CP15 Auxiliary Control Register
  ORR    r0, r0, #(1 <<  1)          // Enable L2 prefetch hint (UNK/WI since r4p1)
  MCR    p15, 0, r0, c1, c0, 1       // Write CP15 Auxiliary Control Register

  // Set Vector Base Address Register (VBAR) to point to this application's vector table
  LDR    R0, =Vectors
  MCR    p15, 0, R0, c12, c0, 0

  // Setup Stack for each exceptional mode
  IMPORT |Image$$FIQ_STACK$$ZI$$Limit|
  IMPORT |Image$$IRQ_STACK$$ZI$$Limit|
  IMPORT |Image$$SVC_STACK$$ZI$$Limit|
  IMPORT |Image$$ABT_STACK$$ZI$$Limit|
  IMPORT |Image$$UND_STACK$$ZI$$Limit|
  IMPORT |Image$$ARM_LIB_STACK$$ZI$$Limit|
  CPS    #0x11
  LDR    SP, =|Image$$FIQ_STACK$$ZI$$Limit|
  CPS    #0x12
  LDR    SP, =|Image$$IRQ_STACK$$ZI$$Limit|
  CPS    #0x13
  LDR    SP, =|Image$$SVC_STACK$$ZI$$Limit|
  CPS    #0x17
  LDR    SP, =|Image$$ABT_STACK$$ZI$$Limit|
  CPS    #0x1B
  LDR    SP, =|Image$$UND_STACK$$ZI$$Limit|
  CPS    #0x1F
  LDR    SP, =|Image$$ARM_LIB_STACK$$ZI$$Limit|

  // Call SystemInit
  IMPORT SystemInit
  BL     SystemInit

  // Unmask interrupts
  CPSIE  if

  // Call __main
  IMPORT __main
  BL     __main
}

/*----------------------------------------------------------------------------
  Default Handler for Exceptions / Interrupts
 *----------------------------------------------------------------------------*/
void Default_Handler(void) {
	while(1);
}

/*----------------------------------------------------------------------------
  Interrupt Handlers
 *----------------------------------------------------------------------------*/
#define SAIC_BASE        0xF803C000
#define AIC_BASE         0xFC020000
#define AIC_IVR          0x10
#define AIC_EOICR        0x38
#define I_BIT            0x80
#define F_BIT            0x40

__ASM void IRQ_Handler (void) {

  /* Save interrupt context on the stack to allow nesting */

  sub     lr, lr, #4
  stmfd   sp!, {lr}
  mrs     lr, SPSR
  stmfd   sp!, {r0, lr}

  /* Write in the IVR to support Protect Mode */

  ldr     lr, =AIC_BASE
  ldr     r0, [r14, #AIC_IVR]
  str     lr, [r14, #AIC_IVR]


  /* Branch to interrupt handler in Supervisor mode */

  msr     CPSR_c, #SVC_MODE
  stmfd   sp!, {r1-r3, r4, r12, lr}

  /* Check for 8-byte alignment and save lr plus a */
  /* word to indicate the stack adjustment used (0 or 4) */

  and     r1, sp, #4
  sub     sp, sp, r1
  stmfd   sp!, {r1, lr}

  blx     r0

  ldmia   sp!, {r1, lr}
  add     sp, sp, r1

  ldmia   sp!, {r1-r3, r4, r12, lr}
  msr     CPSR_c, #IRQ_MODE | I_BIT | F_BIT

  /* Acknowledge interrupt */

  ldr     lr, =AIC_BASE
  str     lr, [r14, #AIC_EOICR]

  /* Restore interrupt context and branch back to calling code */

  ldmia   sp!, {r0, lr}
  msr     SPSR_cxsf, lr
  ldmia   sp!, {pc}^

}

__ASM void FIQ_Handler (void) {

  sub         lr, lr, #4
  stmfd       sp!, {lr}
  stmfd       sp!, {r0}

  /* Write in the IVR to support Protect Mode */

  ldr     lr, =SAIC_BASE
  ldr         r0, [r14, #AIC_IVR]
  str         lr, [r14, #AIC_IVR]

  /* Branch to interrupt handler in Supervisor mode */

  msr         CPSR_c, #SVC_MODE
  stmfd       sp!, { r1-r3, r4, r12, lr}

  blx          r0

  ldmia       sp!, { r1-r3, r4, r12, lr}
  msr         CPSR_c, #FIQ_MODE | I_BIT | F_BIT

  /* Acknowledge interrupt */

  ldr     lr, =SAIC_BASE
  str         lr, [r14, #AIC_EOICR]

  /* Restore interrupt context and branch back to calling code */

  ldmia       sp!, {r0}
  ldmia       sp!, {pc}^

}
