/******************************************************************************
 * @file     SAMA5D2.c
 * @brief    CMSIS Device Header File for SAMA5D2 Device Series
 * @version  V1.00
 * @date     13 Mar 2017
 *
 * @note
 *
 ******************************************************************************/
/*
 * Copyright (c) 2009-2017 ARM Limited. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _SAMA5D2_
#define _SAMA5D2_

/* --------  Configuration of the Cortex-A5 Processor and Core Peripherals  ------- */
#define L2C_310_BASE                          0x00A00000U

#define __CA_REV        0x0000U    /*!< Core revision r0p0                          */
#define __CORTEX_A           5U    /*!< Cortex-A5 Core                              */
#define __FPU_PRESENT        1U    /* FPU present                                   */
#define __GIC_PRESENT        0U    /* GIC present                                   */
#define __TIM_PRESENT        0U    /* TIM present                                   */
#define __L2C_PRESENT        1U    /* L2C present                                   */

#include "core_ca.h"
#include "system_SAMA5D2.h"
#include "sama5d2x.h"
#include "compiler.h"

#define CACHE_ALIGNED __ALIGNED(32U)
#define CPSR_MASK_IRQ                             CPSR_I_Msk
#define CPSR_MASK_FIQ                             CPSR_F_Msk
#define cpsr_set_bits(val)                        __set_CPSR(val)
#define cpsr_clear_bits(mask)                     __set_CPSR(__get_CPSR()&~mask)
#define irq_wait()                                __WFI()
#define cp15_dcache_is_enabled()                  (__get_SCTLR() & (1U << 2U))
#define cp15_icache_invalidate()                  L1C_InvalidateICacheAll()
#define icache_invalidate()                       L1C_InvalidateICacheAll()
#define l2cc_is_enabled()                         (L2C_310->CONTROL & 1U)

__STATIC_INLINE void cp15_dcache_invalidate_region(uint32_t start, uint32_t end) {
  uint32_t mva;
  for (mva = start & ~0x1F; mva < end; mva += 32U)
    L1C_InvalidateDCacheMVA((void*)mva);
}
__STATIC_INLINE void cp15_dcache_clean_region(uint32_t start, uint32_t end) {
  uint32_t mva;
  for (mva = start & ~0x1F; mva < end; mva += 32U)
    L1C_InvalidateDCacheMVA((void*)mva);
}
__STATIC_INLINE void l2cc_invalidate_region(uint32_t start, uint32_t end)
{
  uint32_t current = start & ~0x1F;
  if (l2cc_is_enabled()) {
    while (current <= end) {
      L2C_InvPa((void*)current);
      current += 32;
    }
    L2C_InvPa((void*)end);
  }
}
__STATIC_INLINE void l2cc_clean_region(uint32_t start, uint32_t end)
{
  uint32_t current = start & ~0x1F;
  if (l2cc_is_enabled()) {
    while (current <= end) {
      L2C_CleanPa((void*)current);
      current += 32;
    }
    L2C_CleanPa((void*)end);
  }
}
__STATIC_INLINE void cache_invalidate_region(void *start, uint32_t length)
{
  uint32_t start_addr = (uint32_t)start;
  uint32_t end_addr = start_addr + length;

  if (cp15_dcache_is_enabled()) {
    cp15_dcache_invalidate_region(start_addr, end_addr);
    if (l2cc_is_enabled())
      l2cc_invalidate_region(start_addr, end_addr);
  }
}
__STATIC_INLINE void cache_clean_region(const void *start, uint32_t length)
{
  uint32_t start_addr = (uint32_t)start;
  uint32_t end_addr = start_addr + length;

  if (cp15_dcache_is_enabled()) {
    cp15_dcache_clean_region(start_addr, end_addr);
    if (l2cc_is_enabled())
      l2cc_clean_region(start_addr, end_addr);
  }
}

#endif /* _SAMA5D2_ */
