/**************************************************************************//**
 * @file     irq_ctrl_aic.c
 * @brief    Interrupt controller handling implementation for Atmel AIC
 * @version  V1.0.0
 * @date     05. July 2017
 ******************************************************************************/
/*
 * Copyright (c) 2017 ARM Limited. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stddef.h>

#include "RTE_Components.h"
#include CMSIS_device_header
#include "chip.h"

#include "irq_ctrl.h"


/// Initialize interrupt controller.
__WEAK int32_t IRQ_Initialize (void) {
  int32_t i;

  /* Disable interrupts at core level */
  __disable_irq();
  __disable_fiq();

  /* Disable all interrupts and clear pending flags */
  for (i = 1; i < ID_PERIPH_COUNT; i++)
  {
    AIC->AIC_SSR = i;
    AIC->AIC_IDCR = AIC_IDCR_INTD;
    AIC->AIC_ICCR = AIC_ICCR_INTCLR;
  }

  /* Perform 8 IT acknowledge (write any value in EOICR) */
  for (i = 0; i < 8; i++) AIC->AIC_EOICR = 0;

  /* Clear default handlers */
  for (i = 0; i < ID_PERIPH_COUNT; i++)
  {
    AIC->AIC_SSR = i;
    AIC->AIC_SVR = 0;
  }
  /* Spurious interrupt handler */
  AIC->AIC_SPU = 0;

  /* Redirect all interrupts to non-secure AIC */
  SFR->SFR_AICREDIR = (SFR_AICREDIR_AICREDIRKEY((uint32_t)(AICREDIR_KEY)) ^ SFR->SFR_SN1) | SFR_AICREDIR_NSAIC;

  /* Set debug protection mode */
  AIC->AIC_DCR |= AIC_DCR_PROT;

  /* Enable interrupts at core level */
  __enable_irq();
  __enable_fiq();
  return (0);
}


/// Register interrupt handler.
__WEAK int32_t IRQ_SetHandler (IRQn_ID_t irqn, IRQHandler_t handler) {
	if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (-1);
	AIC->AIC_SSR = AIC_SSR_INTSEL(irqn);
  AIC->AIC_SVR = (uint32_t)handler;
  return (0);
}


/// Get the registered interrupt handler.
__WEAK IRQHandler_t IRQ_GetHandler (IRQn_ID_t irqn) {
  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return ((IRQHandler_t)0);
  AIC->AIC_SSR = AIC_SSR_INTSEL(irqn);
  return ((IRQHandler_t)AIC->AIC_SVR);
}


/// Enable interrupt.
__WEAK int32_t IRQ_Enable (IRQn_ID_t irqn) {
  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (-1);
  AIC->AIC_SSR = AIC_SSR_INTSEL(irqn);
  AIC->AIC_IECR = AIC_IECR_INTEN;
  return (0);
}


/// Disable interrupt.
__WEAK int32_t IRQ_Disable (IRQn_ID_t irqn) {
  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (-1);
  AIC->AIC_SSR = AIC_SSR_INTSEL(irqn);
  AIC->AIC_IDCR = AIC_IDCR_INTD;
  return (0);
}


/// Get interrupt enable state.
__WEAK uint32_t IRQ_GetEnableState (IRQn_ID_t irqn) {
  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (0U);
  AIC->AIC_SSR = AIC_SSR_INTSEL(irqn);
  return (AIC->AIC_IMR);
}


/// Configure interrupt request mode.
__WEAK int32_t IRQ_SetMode (IRQn_ID_t irqn, uint32_t mode) {
  uint32_t cfg = 0;
  uint32_t val;

  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (-1);

  // Check triggering mode
  val = (mode & IRQ_MODE_TRIG_Msk);

  switch(val) {
    case IRQ_MODE_TRIG_EDGE:
    case IRQ_MODE_TRIG_EDGE_FALLING: cfg |= AIC_SMR_SRCTYPE_INT_EDGE_TRIGGERED;  break;
    case IRQ_MODE_TRIG_EDGE_RISING:  cfg |= AIC_SMR_SRCTYPE_EXT_POSITIVE_EDGE;   break;
    case IRQ_MODE_TRIG_LEVEL:
    case IRQ_MODE_TRIG_LEVEL_LOW:    cfg |= AIC_SMR_SRCTYPE_INT_LEVEL_SENSITIVE; break;
    case IRQ_MODE_TRIG_LEVEL_HIGH:   cfg |= AIC_SMR_SRCTYPE_EXT_HIGH_LEVEL;      break;
  }

  AIC->AIC_SSR = AIC_SSR_INTSEL(irqn);
  AIC->AIC_SMR = (AIC->AIC_SMR & ~AIC_SMR_SRCTYPE_Msk) | cfg;
  return (0);
}


/// Get interrupt mode configuration.
__WEAK uint32_t IRQ_GetMode (IRQn_ID_t irqn) {
  uint32_t mode;
  uint32_t val;

  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (IRQ_MODE_ERROR);

  AIC->AIC_SSR = AIC_SSR_INTSEL(irqn);
  val = AIC->AIC_SMR & AIC_SMR_SRCTYPE_Msk;

  switch(val) {
    case AIC_SMR_SRCTYPE_INT_EDGE_TRIGGERED:  mode = IRQ_MODE_TRIG_EDGE;        break;
    case AIC_SMR_SRCTYPE_EXT_POSITIVE_EDGE:   mode = IRQ_MODE_TRIG_EDGE_RISING; break;
    case AIC_SMR_SRCTYPE_INT_LEVEL_SENSITIVE: mode = IRQ_MODE_TRIG_LEVEL;       break;
    case AIC_SMR_SRCTYPE_EXT_HIGH_LEVEL:      mode = IRQ_MODE_TRIG_LEVEL_HIGH;  break;
    default:                                  mode = IRQ_MODE_ERROR;            break;
  }
  return (mode);
}


/// Get ID number of current interrupt request (IRQ).
__WEAK IRQn_ID_t IRQ_GetActiveIRQ (void) {
  IRQHandler_t handler;

  // Read/write interrupt vector register
  handler = (IRQHandler_t)AIC->AIC_IVR;
  AIC->AIC_IVR = (uint32_t)handler;

  // Read current interrupt identifier
  return (AIC->AIC_ISR);
}


/// Get ID number of current fast interrupt request (FIQ).
__WEAK IRQn_ID_t IRQ_GetActiveFIQ (void) {
  // Not supported
  return ((IRQn_ID_t)-1);
}


/// Signal end of interrupt processing.
__WEAK int32_t IRQ_EndOfInterrupt (IRQn_ID_t irqn) {
  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (-1);
  AIC->AIC_EOICR = 0;
  return (0);
}


/// Set interrupt pending flag.
__WEAK int32_t IRQ_SetPending (IRQn_ID_t irqn) {
  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (-1);
  AIC->AIC_SSR = AIC_SSR_INTSEL(irqn);
  AIC->AIC_ISCR = AIC_ISCR_INTSET;
  return (0);
}

/// Get interrupt pending flag.
__WEAK uint32_t IRQ_GetPending (IRQn_ID_t irqn) {
  uint32_t *adr;
  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (0U);
  adr = (uint32_t*)&AIC->AIC_IPR0;
  adr += irqn/32U;
  return (*adr >> (irqn % 32U)) & 1U;
}


/// Clear interrupt pending flag.
__WEAK int32_t IRQ_ClearPending (IRQn_ID_t irqn) {
  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (-1);
  AIC->AIC_SSR = AIC_SSR_INTSEL(irqn);
  AIC->AIC_ICCR = AIC_ICCR_INTCLR;
  return (0);
}


/// Set interrupt priority value.
__WEAK int32_t IRQ_SetPriority (IRQn_ID_t irqn, uint32_t priority) {
  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (-1);
  AIC->AIC_SSR = AIC_SSR_INTSEL(irqn);
  AIC->AIC_SMR = (AIC->AIC_SMR & ~AIC_SMR_PRIOR_Msk) | AIC_SMR_PRIOR(7U-priority);
  return (0);
}


/// Get interrupt priority.
__WEAK uint32_t IRQ_GetPriority (IRQn_ID_t irqn) {
  if ((irqn < 0) || (irqn >= ID_PERIPH_COUNT)) return (IRQ_PRIORITY_ERROR);
  AIC->AIC_SSR = AIC_SSR_INTSEL(irqn);
  return (7U - ((AIC->AIC_SMR & AIC_SMR_PRIOR_Msk) >> AIC_SMR_PRIOR_Pos));
}


/// Set priority masking threshold.
__WEAK int32_t IRQ_SetPriorityMask (uint32_t priority) {
  // Not supported
  return (-1U);
}


/// Get priority masking threshold
__WEAK uint32_t IRQ_GetPriorityMask (void) {
  // Not supported
  return (0U);
}


/// Set priority grouping field split point
__WEAK int32_t IRQ_SetPriorityGroupBits (uint32_t bits) {
  // Not supported
  return (-1);
}


/// Get priority grouping field split point
__WEAK uint32_t IRQ_GetPriorityGroupBits (void) {
  // Not supported
  return (0U);
}
