/**************************************************************************//**
 * @file     os_tick_timer.c
 * @brief    CMSIS OS Tick implementation for OS Timer
 * @version  V1.0.0
 * @date     05. July 2017
 ******************************************************************************/
/*
 * Copyright (c) 2017-2017 ARM Limited. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "os_tick.h"
#include "irq_ctrl.h"

#include "RTE_Components.h"
#include CMSIS_device_header

#include "peripherals/pmc.h"
#include "peripherals/tc.h"

// Define Timer interrupt priority, module, channel and id number
#ifndef TIM_IRQ_PRIORITY
#define TIM_IRQ_PRIORITY           7U
#endif
#ifndef TIM_MODULE
#define TIM_MODULE                TC0
#endif
#ifndef TIM_CHANNEL
#define TIM_CHANNEL                0U
#endif
#ifndef TIM_ID
#define TIM_ID                  0x23U
#endif

// Timer tick frequency
static uint32_t TIM_Clock;

// Timer interrupt pending flag
static uint8_t  TIM_PendIRQ;


// Setup OS Tick.
int32_t OS_Tick_Setup (uint32_t freq, IRQHandler_t handler) {
  uint32_t tc_clks, config, load;

  if (freq == 0U) {
    return (-1);
  }
  
  if (!pmc_is_peripheral_enabled(TIM_ID))
    pmc_configure_peripheral(TIM_ID, NULL, true);

  tc_clks = tc_find_best_clock_source(TIM_MODULE, TIM_CHANNEL, SystemCoreClock);
  config = tc_clks | TC_CMR_WAVE | TC_CMR_WAVSEL_UP_RC;
  tc_configure(TIM_MODULE, TIM_CHANNEL, config);
  TIM_Clock = tc_get_channel_freq(TIM_MODULE, TIM_CHANNEL);

  load = (TIM_Clock / freq) - 1U;
  tc_set_ra_rb_rc(TIM_MODULE, TIM_CHANNEL, NULL, NULL, &load);
  tc_enable_it(TIM_MODULE, TIM_CHANNEL, TC_IER_CPCS);

  TIM_PendIRQ = 0U;

  // Disable corresponding IRQ
  IRQ_Disable     (TIM_ID);
  IRQ_ClearPending(TIM_ID);

  // Set lowest interrupt priority
  IRQ_SetPriority (TIM_ID, TIM_IRQ_PRIORITY);

  // Set edge-triggered IRQ
  IRQ_SetMode (TIM_ID, IRQ_MODE_TRIG_EDGE);

  // Register tick interrupt handler function
  IRQ_SetHandler(TIM_ID, handler);

  // Enable corresponding IRQ
  IRQ_Enable (TIM_ID);

  return (0);
}

/// Enable OS Tick.
__WEAK int32_t  OS_Tick_Enable (void) {

  if (TIM_PendIRQ != 0U) {
	  TIM_PendIRQ = 0U;
    IRQ_SetPending (TIM_ID);
  }

  // Start the counter
  tc_start(TIM_MODULE, TIM_CHANNEL);

  return (0);
}

/// Disable OS Tick.
__WEAK int32_t  OS_Tick_Disable (void) {

  // Stop the counter
  tc_stop(TIM_MODULE, TIM_CHANNEL);

  if (IRQ_GetPending(TIM_ID) != 0) {
    IRQ_ClearPending (TIM_ID);
    TIM_PendIRQ = 1U;
  }

  return (0);
}

// Acknowledge OS Tick IRQ.
int32_t OS_Tick_AcknowledgeIRQ (void) {

  uint32_t status = tc_get_status(TIM_MODULE, TIM_CHANNEL);

  if ((status & TC_SR_CPCS) == TC_SR_CPCS) {
	return (0);
  }
  return (-1);
}

// Get OS Tick IRQ number.
int32_t  OS_Tick_GetIRQn (void) {
  return (TIM_ID);
}

// Get OS Tick clock.
uint32_t OS_Tick_GetClock (void) {
  return TIM_Clock;
}

// Get OS Tick interval.
uint32_t OS_Tick_GetInterval (void) {
  return (TIM_MODULE->TC_CHANNEL[0].TC_RC + 1U);
}

// Get OS Tick count value.
uint32_t OS_Tick_GetCount (void) {
  return  (TIM_MODULE->TC_CHANNEL[0].TC_RC - TIM_MODULE->TC_CHANNEL[0].TC_CV);
}

// Get OS Tick overflow status.
uint32_t OS_Tick_GetOverflow (void) {
  return ((TIM_MODULE->TC_CHANNEL[0].TC_SR & TC_SR_COVFS) ? (1U) : (0U));
}
