/* ----------------------------------------------------------------------------
 *         SAM Software Package License
 * ----------------------------------------------------------------------------
 * Copyright (c) 2015-2016, Atmel Corporation
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 * Modifications Copyright (c) 2017 ARM Germany GmbH. All rights reserved.
 */

/**
 * \file
 *
 * Implementation of memories configuration on board.
 *
 */

/*----------------------------------------------------------------------------
 *        Headers
 *----------------------------------------------------------------------------*/

#include "board.h"
#include "board_timer.h"
#include "trace.h"

#include "irq/irq.h"
#ifdef CONFIG_HAVE_ISC
#include "video/isc.h"
#endif
#include "gpio/pio.h"
#include "peripherals/pmc.h"
#include "board_support.h"

#ifdef CONFIG_HAVE_PMIC_ACT8945A
#include "board_twi.h"
#include "power/act8945a.h"
#endif

/*----------------------------------------------------------------------------
 *        Local types
 *----------------------------------------------------------------------------*/

struct _console_cfg {
	void* addr;
	const struct _pin pins[2];
};

/*----------------------------------------------------------------------------
 *        Local constants
 *----------------------------------------------------------------------------*/

static const char* board_name = BOARD_NAME;

/*----------------------------------------------------------------------------
 *        Local variables
 *----------------------------------------------------------------------------*/
#ifdef CONFIG_HAVE_PMIC_ACT8945A
static struct _act8945a act8945a = {
	.bus = BOARD_ACT8945A_TWI_BUS,
	.addr = BOARD_ACT8945A_TWI_ADDR,
	.desc = {
		.pin_chglev = BOARD_ACT8945A_PIN_CHGLEV,
		.pin_irq = BOARD_ACT8945A_PIN_IRQ,
		.pin_lbo = BOARD_ACT8945A_PIN_LBO
	}
};

static bool act8945a_initialized = false;
#endif /* CONFIG_HAVE_PMIC_ACT8945A */

/*----------------------------------------------------------------------------
 *        Exported functions
 *----------------------------------------------------------------------------*/
const char* get_board_name(void)
{
	return board_name;
}

void board_cfg_clocks(void)
{
	struct _pmc_plla_cfg plla_config = {
		.mul = 82,
		.div = 1,
		.count = 0x3f,
	};
	pmc_switch_mck_to_slck();
	pmc_set_mck_h32mxdiv(true);
	pmc_set_mck_plladiv2(true);
	pmc_set_mck_prescaler(PMC_MCKR_PRES_CLOCK);
	pmc_set_mck_divider(PMC_MCKR_MDIV_EQ_PCK);
	pmc_disable_plla();
	pmc_select_external_osc();
	pmc_configure_plla(&plla_config);
	pmc_set_mck_divider(PMC_MCKR_MDIV_PCK_DIV3);
	pmc_set_mck_prescaler(PMC_MCKR_PRES_CLOCK);
	pmc_switch_mck_to_pll();
}

void board_cfg_lowlevel(bool clocks, bool ddram, bool mmu)
{
	/* Disable all PIO interrupts */
	pio_reset_all_it();

	/* Set the external clock values */
	pmc_set_oscillators(BOARD_SLOW_CLOCK_EXT_OSC, BOARD_MAIN_CLOCK_EXT_OSC);

	if (clocks) {
		/* Configure system clocks */
		board_cfg_clocks();
	}

	/* Setup default interrupt handlers */
	irq_initialize();

	/* Configure system timer */
	board_cfg_timer();
}

void board_restore_pio_reset_state(void)
{
	int i;

	/* all pins, excluding JTAG and NTRST */
	struct _pin pins[] = {
		{ PIO_GROUP_A, 0xFFFFFFFF, PIO_INPUT, PIO_PULLUP },
		{ PIO_GROUP_B, 0xFFFFFFFF, PIO_INPUT, PIO_PULLUP },
		{ PIO_GROUP_C, 0xFFFFFFFF, PIO_INPUT, PIO_PULLUP },
		{ PIO_GROUP_D, 0xFFF83FFF, PIO_INPUT, PIO_PULLUP },
	};

	/* For low_power_mode example, power consumption results can be affected
	* by IOs setting. To generate power consumption numbers in datasheet,
	* most IOs must be disconnected from external devices just like on
	* VB board. Then putting IOs to reset state are OK.
	* On SAMA5D2-XULT board, please put IOs to output state as below.
	*/
	//struct _pin pins[] = {
	//	{ PIO_GROUP_A, 0xFFFFFFFF, PIO_OUTPUT, PIO_PULLUP },
	//	{ PIO_GROUP_B, 0xFFFFFFFF, PIO_OUTPUT, PIO_PULLUP },
	//	{ PIO_GROUP_C, 0xFFFFFFFF, PIO_OUTPUT, PIO_PULLUP },
	//	{ PIO_GROUP_D, 0xFFF83FFF, PIO_OUTPUT, PIO_PULLUP },
	//};

	pio_configure(pins, ARRAY_SIZE(pins));
	for (i = 0; i < ARRAY_SIZE(pins); i++)
		pio_clear(&pins[i]);
}

void board_save_misc_power(void)
{
	int i;

	/* disable USB clock */
	pmc_disable_upll_clock();
	pmc_disable_upll_bias();

	/* Disable audio clock */
	pmc_disable_audio();

	/* disable system clocks */
	pmc_disable_system_clock(PMC_SYSTEM_CLOCK_DDR);
	pmc_disable_system_clock(PMC_SYSTEM_CLOCK_LCD);
	pmc_disable_system_clock(PMC_SYSTEM_CLOCK_UHP);
	pmc_disable_system_clock(PMC_SYSTEM_CLOCK_UDP);
	pmc_disable_system_clock(PMC_SYSTEM_CLOCK_PCK0);
	pmc_disable_system_clock(PMC_SYSTEM_CLOCK_PCK1);
	pmc_disable_system_clock(PMC_SYSTEM_CLOCK_PCK2);
#ifdef CONFIG_HAVE_ISC
	pmc_disable_system_clock(PMC_SYSTEM_CLOCK_ISC);
#endif

	/* disable all peripheral clocks except PIOA for JTAG, serial debug port */
	for (i = ID_PIT; i < ID_PERIPH_COUNT; i++) {
		if (i == ID_PIOA)
			continue;
		pmc_disable_peripheral(i);
	}
}

#ifdef CONFIG_HAVE_NAND_FLASH
void board_cfg_nand_flash(void)
{
#if defined(BOARD_NANDFLASH_PINS) && defined(BOARD_NANDFLASH_BUS_WIDTH)
	const struct _pin pins_nandflash[] = BOARD_NANDFLASH_PINS;
	pio_configure(pins_nandflash, ARRAY_SIZE(pins_nandflash));
	board_cfg_matrix_for_nand();
	smc_nand_configure(BOARD_NANDFLASH_BUS_WIDTH);
#else
	trace_fatal("Cannot configure NAND: target board has no NAND definitions!");
#endif
}
#endif /* CONFIG_HAVE_NAND_FLASH */

#ifdef CONFIG_HAVE_SDMMC
bool board_cfg_sdmmc(uint32_t periph_id)
{
/* mask for board capabilities defines: voltage, slot type and 8-bit support */
#define CAPS0_MASK (SDMMC_CA0R_V33VSUP | SDMMC_CA0R_V30VSUP | \
                    SDMMC_CA0R_V18VSUP | SDMMC_CA0R_SLTYPE_Msk | \
                    SDMMC_CA0R_ED8SUP)

	switch (periph_id) {
#ifdef SDMMC0
	case ID_SDMMC0:
	{
#if defined(BOARD_SDMMC0_CAPS0) && defined(BOARD_SDMMC0_PINS)
		struct _pin pins[] = BOARD_SDMMC0_PINS;
		uint32_t caps0 = BOARD_SDMMC0_CAPS0;

#ifdef CONFIG_BOARD_SAMA5D2_XPLAINED
		struct _pin vsel_pin = PIN_SDMMC0_VDDSEL_IOS1;
		int vsel_ix;

#ifndef SDMMC_TRIM_LOW_VOLTAGE
		/* Set PMIC output 7 to 1.8V (VDDSDHC1V8) */
		if (!act8945a_initialized) {
			board_cfg_pmic();
			if (!act8945a_initialized)
				return false;
		}
		if (!act8945a_set_regulator_voltage(&act8945a, 7, 1800))
			return false;
		if (!act8945a_enable_regulator(&act8945a, 7, true))
			return false;
#endif

		/* The PIOs of SDMMC0 normally include SDMMC0_VDDSEL. On regular
		 * SAMA5D2-XULT, the SDMMC0_VDDSEL line has a pull-down resistor
		 * hence at power-on time VCCQ is 3.3V. In this default config
		 * we enable SDMMC0_VDDSEL, which can switch VCCQ to 1.8V.
		 * Changing VCCQ on the fly is required with UHS-I SD cards. It
		 * is illegal with e.MMC devices. Detect if the board has been
		 * modified to supply 1.8V VCCQ at power-on time.
		 * First, search for the SDMMC0_VDDSEL PIO on this board. */
		for (vsel_ix = 0; vsel_ix < ARRAY_SIZE(pins); vsel_ix++)
			if (pins[vsel_ix].mask == vsel_pin.mask
			    && pins[vsel_ix].group == vsel_pin.group)
				break;
		if (vsel_ix < ARRAY_SIZE(pins)) {
			/* Second, sense whether the SDMMC0_VDDSEL line is
			 * pulled up or down */
			vsel_pin.type = PIO_INPUT;
			vsel_pin.attribute = PIO_PULLUP;
			pio_configure(&vsel_pin, 1);
			if (pio_get(&vsel_pin)) {
				/* The line is pulled up => at power-on time
				 * VCCQ is 1.8V. If the SDMMC0_VDDSEL function
				 * was enabled, then everytime SRR:SWRSTALL was
				 * triggered, VCCQ would switch to 3.3V. */
				pins[vsel_ix].type = PIO_OUTPUT_1;
				pins[vsel_ix].attribute = PIO_DEFAULT;
				/* Deviation from the specification: we use the
				 * Voltage Support capabilities to indicate the
				 * supported signaling levels (VCCQ). */
				caps0 &= ~SDMMC_CA0R_V33VSUP;
				caps0 &= ~SDMMC_CA0R_V30VSUP;
				caps0 |= SDMMC_CA0R_V18VSUP;
			}
		}
#endif
		/* Program capabilities for SDMMC0 */
		sdmmc_set_capabilities(SDMMC0, caps0, CAPS0_MASK, 0, 0);

		/* Configure SDMMC0 pins */
		pio_configure(pins, ARRAY_SIZE(pins));
		return true;
#else
		trace_fatal("Cannot configure SDMMC0: target board has no SDMMC0 definitions!");
		return false;
#endif
	}
#endif
	case ID_SDMMC1:
	{
#if defined(BOARD_SDMMC1_CAPS0) && defined(BOARD_SDMMC1_PINS)
		const struct _pin pins[] = BOARD_SDMMC1_PINS;
		uint32_t caps0 = BOARD_SDMMC1_CAPS0;

		/* Program capabilities for SDMMC1 */
		sdmmc_set_capabilities(SDMMC1, caps0, CAPS0_MASK, 0, 0);

		/* Configure SDMMC1 pins */
		pio_configure(pins, ARRAY_SIZE(pins));
		return true;
#else
		trace_fatal("Cannot configure SDMMC1: target board has no SDMMC1 definitions!");
		return false;
#endif
	}
	default:
		return false;
	}

#undef CAPS0_MASK
}
#endif /* CONFIG_HAVE_SDMMC */

void board_cfg_pmic(void)
{
#ifdef CONFIG_HAVE_PMIC_ACT8945A
	if (act8945a_initialized)
		return;

	if (!act8945a_configure(&act8945a))
		goto Fail;
#if defined(CONFIG_BOARD_SAMA5D2_XPLAINED) || defined(CONFIG_BOARD_SAMA5D2_PTC_ENGI)
	/* Set PMIC output 6 to 2.5V (VDD_LED) */
	if (!act8945a_set_regulator_voltage(&act8945a, 6, 2500))
		goto Fail;
	if (!act8945a_enable_regulator(&act8945a, 6, true))
		goto Fail;
#endif
	act8945a_initialized = true;
	return;

Fail:
	trace_error("Error initializing ACT8945A PMIC\r\n");
#endif
}
